@@ -60,16 +60,16 @@
  * The #AutoarExtractor object is used to automatically extract files and
  * directories from an archive. By default, it will only create one file or
  * directory in the output directory. This is done to avoid clutter on the
- * user's output directory. If the archive contains only one file, the file
- * will be extracted to the output directory. If the archive has more than one
- * file, the files will be extracted in a directory having the same name as the
- * archive, except the extension. It is also possible to just extract all files
- * to the output directory (note that this will not perform any checks) by
- * using autoar_extractor_set_output_is_dest().
+ * user's output directory. If the archive contains only one file with the same
+ * name as the source archive without the extension, the file will be extracted
+ * to the output directory. Otherwise the files will be extracted in a directory
+ * having the same name as the source archive, except the extension. It is also
+ * possible to just extract all files to the output directory (note that this
+ * will not perform any checks) by using autoar_extractor_set_output_is_dest().
 
  * #AutoarExtractor will not attempt to solve any name conflicts. If the
  * destination directory already exists, it will proceed normally. If the
- * destionation directory cannot be created, it will fail with an error.
+ * destination directory cannot be created, it will fail with an error.
  * It is possible however to change the destination, when
  * #AutoarExtractor::decide-destination is emitted. The signal provides the decided
  * destination and the list of files to be extracted. The signal also allows a
@@ -289,8 +289,9 @@ autoar_extractor_get_source_file (AutoarExtractor *self)
  * autoar_extractor_get_output_file:
  * @self: an #AutoarExtractor
  *
- * This function is similar to autoar_extractor_get_output(), except for the
- * return value is a #GFile.
+ * Gets the #GFile object which represents the output directory of extracted
+ * file or directory, or the extracted file or directory itself if you set
+ * #AutoarExtractor:output-is-dest on the returned object.
  *
  * Returns: (transfer none): a #GFile
  **/
@@ -368,8 +369,8 @@ autoar_extractor_get_completed_files (AutoarExtractor *self)
  *
  * See autoar_extractor_set_output_is_dest().
  *
- * Returns: %TRUE if #AutoarExtractor:output is the location of extracted file
- * or directory
+ * Returns: %TRUE if #AutoarExtractor:output-file is the destination for
+ * extracted files
  **/
 gboolean
 autoar_extractor_get_output_is_dest (AutoarExtractor *self)
@@ -413,16 +414,17 @@ autoar_extractor_get_notify_interval (AutoarExtractor *self)
 /**
  * autoar_extractor_set_output_is_dest:
  * @self: an #AutoarExtractor
- * @output_is_dest: %TRUE if the location of the extracted directory or file
- * has been already decided
+ * @output_is_dest: %TRUE if #AutoarExtractor:output-file is the destination for
+ * extracted files
  *
  * By default #AutoarExtractor:output-is-dest is set to %FALSE, which means
- * only one file or directory will be generated. The destination is internally
- * determined by analyzing the contents of the archive. If this is not wanted,
- * #AutoarExtractor:output-is-dest can be set to %TRUE, which will make
- * #AutoarExtractor:output the destination for extracted files. In any case, the
- * destination will be notified via #AutoarExtractor::decide-destination, when
- * it is possible to set a new destination.
+ * only one file or directory will be created in #AutoarExtractor:output-file.
+ * The destination is internally determined by analyzing the contents of the
+ * archive. If this is not wanted, #AutoarExtractor:output-is-dest can be set to
+ * %TRUE, which will make #AutoarExtractor:output-file the destination for
+ * extracted files. In any case, the destination will be notified via
+ * #AutoarExtractor::decide-destination, when it is possible to set a new
+ * destination.
  *
  * #AutoarExtractor will attempt to create the destination regardless to whether
  * its path was internally decided or not.
@@ -778,13 +780,15 @@ autoar_extractor_signal_conflict (AutoarExtractor  *self,
                                   GFile            *file,
                                   GFile           **new_file)
 {
-  AutoarConflictAction action = AUTOAR_CONFLICT_OVERWRITE;
+  AutoarConflictAction action = AUTOAR_CONFLICT_UNHANDLED;
 
   autoar_common_g_signal_emit (self, self->in_thread,
                                autoar_extractor_signals[CONFLICT], 0,
                                file,
                                new_file,
                                &action);
+  if (action == AUTOAR_CONFLICT_UNHANDLED)
+    return AUTOAR_CONFLICT_SKIP;
 
   if (*new_file) {
     g_autofree char *previous_path;
@@ -868,70 +872,35 @@ autoar_extractor_get_common_prefix (GList *files,
   return prefix;
 }
 
-static gboolean
-is_valid_filename (GFile *file, GFile *destination)
-{
-  g_autoptr (GFile) parent = NULL;
-  g_autoptr (GFileInfo) info = NULL;
-  g_autoptr (GError) error = NULL;
-
-  if (g_file_equal (file, destination))
-    return TRUE;
-
-  if (!g_file_has_prefix (file, destination))
-    return FALSE;
-
-  /* Resolve symbolic link ancestors to confirm file is actually inside destination. */
-  parent = g_file_get_parent (file);
-  info = g_file_query_info (parent,
-                            G_FILE_ATTRIBUTE_STANDARD_IS_SYMLINK ","
-                            G_FILE_ATTRIBUTE_STANDARD_SYMLINK_TARGET,
-                            G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS,
-                            NULL,
-                            &error);
-
-  /* The parent directories don't have to be created yet. */
-  if (error && !g_error_matches (error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND))
-    return FALSE;
-
-  if (info && g_file_info_get_is_symlink (info)) {
-    g_autoptr (GFile) cwd = NULL;
-    const gchar *target;
-
-    target = g_file_info_get_symlink_target (info);
-    if (g_path_is_absolute (target))
-      return FALSE;
-
-    cwd = g_file_get_parent (parent);
-    g_object_unref (parent);
-    parent = g_file_resolve_relative_path (cwd, target);
-  }
-
-  /* Climb up the path to resolve every symbolic link ancestor found */
-  return is_valid_filename (parent, destination);
-}
-
 static GFile*
 autoar_extractor_do_sanitize_pathname (AutoarExtractor *self,
                                        const char      *pathname_bytes)
 {
   GFile *extracted_filename;
   gboolean valid_filename;
-  g_autofree char *sanitized_pathname = NULL;
+  g_autofree char *sanitized_pathname;
   g_autofree char *utf8_pathname;
+  GFile *destination;
+
+  /* Use output_file when called from autoar_extractor_step_scan_toplevel(). */
+  destination = (self->destination_dir != NULL) ? self->destination_dir : self->output_file;
 
   utf8_pathname = autoar_common_get_utf8_pathname (pathname_bytes);
-  extracted_filename = g_file_get_child (self->destination_dir,
+  extracted_filename = g_file_get_child (destination,
                                          utf8_pathname ?  utf8_pathname : pathname_bytes);
 
-  valid_filename = is_valid_filename (extracted_filename, self->destination_dir);
+  valid_filename =
+    g_file_equal (extracted_filename, destination) ||
+    g_file_has_prefix (extracted_filename, destination);
+
   if (!valid_filename) {
-    g_warning ("autoar_extractor_do_sanitize_pathname: %s is outside of the destination dir",
-                g_file_peek_path (extracted_filename));
+    g_autofree char *basename;
+
+    basename = g_file_get_basename (extracted_filename);
 
     g_object_unref (extracted_filename);
 
-    return NULL;
+    extracted_filename = g_file_get_child (destination, basename);
   }
 
   if (self->prefix != NULL && self->new_prefix != NULL) {
@@ -955,92 +924,42 @@ autoar_extractor_do_sanitize_pathname (AutoarExtractor *self,
   return extracted_filename;
 }
 
-static gboolean
-autoar_extractor_check_file_conflict (GFile  *file,
+/* The function checks @file for conflicts with already existing files on the
+ * disk. It also recursively checks parents of @file to be sure it is directory.
+ * It doesn't follow symlinks, so symlinks in parents are also considered as
+ * conflicts even though they point to directory. It returns #GFile object for
+ * the file, which cause the conflict (so @file, or some of its parents). If
+ * there aren't any conflicts, NULL is returned.
+ */
+static GFile *
+autoar_extractor_check_file_conflict (AutoarExtractor *self,
+                                      GFile  *file,
                                       mode_t  extracted_filetype)
 {
   GFileType file_type;
-  gboolean conflict = FALSE;
+  g_autoptr (GFile) parent = NULL;
 
   file_type = g_file_query_file_type (file,
-                                      G_FILE_QUERY_INFO_NONE,
+                                      G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS,
                                       NULL);
-  /* If there is no file with the given name, there will be no conflict */
-  if (file_type == G_FILE_TYPE_UNKNOWN) {
-    return FALSE;
+
+  /* It is a conflict if the file already exists with an exception for already
+   * existing directories.
+   */
+  if (file_type != G_FILE_TYPE_UNKNOWN &&
+      (file_type != G_FILE_TYPE_DIRECTORY ||
+       extracted_filetype != AE_IFDIR)) {
+    return g_object_ref (file);
   }
 
-  switch (extracted_filetype) {
-    case AE_IFDIR:
-      break;
-    case AE_IFREG:
-    case AE_IFLNK:
-#if defined HAVE_MKFIFO || defined HAVE_MKNOD
-    case AE_IFIFO:
-#endif
-#ifdef HAVE_MKNOD
-    case AE_IFSOCK:
-    case AE_IFBLK:
-    case AE_IFCHR:
-#endif
-      conflict = TRUE;
-      break;
-    default:
-      break;
+  if ((self->new_prefix && g_file_equal (self->new_prefix, file)) ||
+      (!self->new_prefix && g_file_equal (self->destination_dir, file))) {
+    return NULL;
   }
 
-  return conflict;
-}
-
-static gboolean
-autoar_extractor_delete_file_recursively (AutoarExtractor *self,
-                                          GFile *file)
-{
-  gboolean success;
-  g_autoptr (GError) error = NULL;
-
-  do {
-    g_autoptr (GFileEnumerator) enumerator = NULL;
-
-    success = g_file_delete (file, self->cancellable, &error);
-    if (success || !g_error_matches (error, G_IO_ERROR, G_IO_ERROR_NOT_EMPTY)) {
-      break;
-    }
-
-    g_clear_error (&error);
-    enumerator = g_file_enumerate_children (file,
-                                            G_FILE_ATTRIBUTE_STANDARD_NAME,
-                                            G_FILE_QUERY_INFO_NONE,
-                                            self->cancellable, &error);
-    if (enumerator) {
-      GFileInfo *info;
-
-      success = TRUE;
-
-      info = g_file_enumerator_next_file (enumerator,
-                                          self->cancellable,
-                                          &error);
-      while (info != NULL) {
-        g_autoptr (GFile) child = NULL;
-
-        child = g_file_enumerator_get_child (enumerator, info);
-
-        success = success && autoar_extractor_delete_file_recursively (self, child);
-        g_object_unref (info);
-
-        info = g_file_enumerator_next_file (enumerator,
-                                            self->cancellable,
-                                            &error);
-      }
-    }
-
-    if (error != NULL) {
-      success = FALSE;
-    }
-
-  } while (success);
-
-  return success;
+  /* Check also parents for conflict to be sure it is directory. */
+  parent = g_file_get_parent (file);
+  return autoar_extractor_check_file_conflict (self, parent, AE_IFDIR);
 }
 
 static void
@@ -1199,12 +1118,10 @@ autoar_extractor_do_write_entry (AutoarExtractor      *self,
 
         g_debug ("autoar_extractor_do_write_entry: case REG");
 
-        ostream = (GOutputStream*)g_file_replace (dest,
-                                                  NULL,
-                                                  FALSE,
-                                                  G_FILE_CREATE_NONE,
-                                                  self->cancellable,
-                                                  &(self->error));
+        ostream = (GOutputStream*)g_file_create (dest,
+                                                 G_FILE_CREATE_NONE,
+                                                 self->cancellable,
+                                                 &(self->error));
         if (self->error != NULL) {
           g_object_unref (info);
           return;
@@ -1242,7 +1159,6 @@ autoar_extractor_do_write_entry (AutoarExtractor      *self,
             }
 
             if (r == ARCHIVE_FAILED) {
-              autoar_extractor_delete_file_recursively (self, self->destination_dir);
               if (self->error == NULL) {
                 self->error = g_error_new (AUTOAR_EXTRACTOR_ERROR,
                                            INCORRECT_PASSPHRASE_ERRNO,
@@ -1293,11 +1209,17 @@ autoar_extractor_do_write_entry (AutoarExtractor      *self,
       }
       break;
     case AE_IFLNK:
-      g_debug ("autoar_extractor_do_write_entry: case LNK");
-      g_file_make_symbolic_link (dest,
-                                 archive_entry_symlink (entry),
-                                 self->cancellable,
-                                 &(self->error));
+      {
+        const char *symlink_path = archive_entry_symlink (entry);
+
+        g_debug ("autoar_extractor_do_write_entry: case LNK, %s => %s",
+                 g_file_peek_path (dest), symlink_path);
+
+        g_file_make_symbolic_link (dest,
+                                   symlink_path,
+                                   self->cancellable,
+                                   &(self->error));
+      }
       break;
     /* FIFOs, sockets, block files, character files are not important
      * in the regular archives, so errors are not fatal. */
@@ -1401,8 +1323,8 @@ autoar_extractor_class_init (AutoarExtractorClass *klass)
 
   g_object_class_install_property (object_class, PROP_SOURCE_FILE,
                                    g_param_spec_object ("source-file",
-                                                        "Source archive GFile",
-                                                        "The archive GFile to be extracted",
+                                                        "Source archive",
+                                                        "The #GFile of the source archive that will be extracted",
                                                         G_TYPE_FILE,
                                                         G_PARAM_READWRITE |
                                                         G_PARAM_CONSTRUCT_ONLY |
@@ -1410,8 +1332,8 @@ autoar_extractor_class_init (AutoarExtractorClass *klass)
 
   g_object_class_install_property (object_class, PROP_OUTPUT_FILE,
                                    g_param_spec_object ("output-file",
-                                                        "Output directory GFile",
-                                                        "Output directory GFile of extracted archives",
+                                                        "Output file",
+                                                        "The #GFile of the directory where the files will be extracted",
                                                         G_TYPE_FILE,
                                                         G_PARAM_READWRITE |
                                                         G_PARAM_CONSTRUCT_ONLY |
@@ -1452,7 +1374,7 @@ autoar_extractor_class_init (AutoarExtractorClass *klass)
   g_object_class_install_property (object_class, PROP_OUTPUT_IS_DEST,
                                    g_param_spec_boolean ("output-is-dest",
                                                          "Output is destination",
-                                                         "Whether output direcotry is used as destination",
+                                                         "Whether #AutoarExtractor:output-file is used as destination",
                                                          FALSE,
                                                          G_PARAM_READWRITE |
                                                          G_PARAM_CONSTRUCT |
@@ -1498,12 +1420,12 @@ autoar_extractor_class_init (AutoarExtractorClass *klass)
 /**
  * AutoarExtractor::decide-destination:
  * @self: the #AutoarExtractor
- * @destination: the location where files will be extracted
- * @files: the list of files to be extracted. All have @destination as their
-           common prefix
+ * @destination: a #GFile for the location where files will be extracted
+ * @files: the list of #GFile objects to be extracted. All have @destination as
+ *         their common prefix
  *
- * Returns: (transfer full): a new destination that will overwrite the previous
- *                           one, or %NULL if this is not wanted
+ * Returns: (transfer full): #GFile for location that will overwrite
+ *                           @destination, or %NULL if this is not wanted
  *
  * This signal is emitted when the path of the destination is determined. It is
  * useful for solving name conflicts or for setting a new destination, based on
@@ -1526,7 +1448,7 @@ autoar_extractor_class_init (AutoarExtractorClass *klass)
  * @completed_size: bytes has been written to disk
  * @completed_files: number of files have been written to disk
  *
- * This signal is used to report progress of creating archives.
+ * This signal is used to report progress of extraction.
  **/
   autoar_extractor_signals[PROGRESS] =
     g_signal_new ("progress",
@@ -1542,13 +1464,14 @@ autoar_extractor_class_init (AutoarExtractorClass *klass)
 /**
  * AutoarExtractor::conflict:
  * @self: the #AutoarExtractor
- * @file: the file that caused a conflict
- * @new_file: an address to store the new destination for a conflict file
+ * @file: a #GFile for the file that caused a conflict
+ * @new_file: a #GFile for the new destination of @file
  *
- * Returns: the action to be performed by #AutoarExtractor
+ * Returns: the #AutoarConflictAction to be performed by #AutoarExtractor
  *
  * This signal is used to report and offer the possibility to solve name
- * conflicts when extracting files.
+ * conflicts when extracting files. If it is not handled, the @file will be
+ * skipped.
  **/
   autoar_extractor_signals[CONFLICT] =
     g_signal_new ("conflict",
@@ -1669,10 +1592,8 @@ autoar_extractor_init (AutoarExtractor *self)
 
 /**
  * autoar_extractor_new:
- * @source_file: source archive
- * @output_file: output directory of extracted file or directory, or the
- * file name of the extracted file or directory itself if you set
- * #AutoarExtractor:output-is-dest on the returned object
+ * @source_file: a #GFile for the source archive
+ * @output_file: a #GFile for the directory where the files will be extracted
  *
  * Create a new #AutoarExtractor object.
  *
@@ -1750,22 +1671,31 @@ autoar_extractor_step_scan_toplevel (AutoarExtractor *self)
     }
 
     if (self->use_raw_format) {
-      pathname = autoar_common_get_basename_remove_extension (g_file_get_path(self->source_file));
+      pathname = autoar_common_get_basename_remove_extension (g_file_peek_path (self->source_file));
       g_debug ("autoar_extractor_step_scan_toplevel: %d: raw pathname = %s",
                self->total_files, pathname);
     } else {
+      const char *symlink_pathname;
+      const char *hardlink_pathname;
+
       pathname = archive_entry_pathname (entry);
       utf8_pathname = autoar_common_get_utf8_pathname (pathname);
+      symlink_pathname = archive_entry_symlink (entry);
+      hardlink_pathname = archive_entry_hardlink (entry);
 
-      g_debug ("autoar_extractor_step_scan_toplevel: %d: pathname = %s %s%s",
+      g_debug ("autoar_extractor_step_scan_toplevel: %d: pathname = %s%s%s%s%s%s%s",
                self->total_files, pathname,
-               utf8_pathname ? "utf8 pathname = " :"",
-               utf8_pathname ? utf8_pathname : "");
+               utf8_pathname ? " utf8 pathname = " : "",
+               utf8_pathname ? utf8_pathname : "",
+               symlink_pathname ? " symlink = " : "",
+               symlink_pathname ? symlink_pathname : "",
+               hardlink_pathname ? " hardlink = " : "",
+               hardlink_pathname ? hardlink_pathname : "");
     }
     self->files_list =
       g_list_prepend (self->files_list,
-                      g_file_get_child (self->output_file,
-                                        utf8_pathname ? utf8_pathname : pathname));
+                      autoar_extractor_do_sanitize_pathname (self,
+                                                             utf8_pathname ? utf8_pathname : pathname));
     self->total_files++;
     self->total_size += archive_entry_size (entry);
     archive_read_data_skip (a);
@@ -1804,18 +1734,6 @@ autoar_extractor_step_scan_toplevel (AutoarExtractor *self)
 
   self->files_list = g_list_reverse (self->files_list);
 
-  self->prefix =
-    autoar_extractor_get_common_prefix (self->files_list,
-                                        self->output_file);
-
-  if (self->prefix != NULL) {
-    g_autofree char *path_prefix;
-
-    path_prefix = g_file_get_path (self->prefix);
-    g_debug ("autoar_extractor_step_scan_toplevel: pathname_prefix = %s",
-             path_prefix);
-  }
-
   autoar_extractor_signal_scanned (self);
 }
 
@@ -1831,6 +1749,8 @@ autoar_extractor_step_set_destination (AutoarExtractor *self)
     return;
   }
 
+  self->prefix = autoar_extractor_get_common_prefix (self->files_list,
+                                                     self->output_file);
   if (self->prefix != NULL) {
     /* We must check if the archive and the prefix have the same name (without
      * the extension). If they do, then the destination should be the output
@@ -1839,6 +1759,9 @@ autoar_extractor_step_set_destination (AutoarExtractor *self)
     g_autofree char *prefix_name;
     g_autofree char *prefix_name_no_ext;
 
+    g_debug ("autoar_extractor_step_set_destination: pathname_prefix = %s",
+             g_file_peek_path (self->prefix));
+
     prefix_name = g_file_get_basename (self->prefix);
     prefix_name_no_ext = autoar_common_get_basename_remove_extension (prefix_name);
 
@@ -1905,19 +1828,18 @@ autoar_extractor_step_decide_destination (AutoarExtractor *self)
     }
   }
 
-  destination_name = g_file_get_path (self->new_prefix != NULL ?
-                                      self->new_prefix :
-                                      self->destination_dir);
+  new_destination = (self->new_prefix != NULL) ? self->new_prefix : self->destination_dir;
+  destination_name = g_file_get_path (new_destination);
   g_debug ("autoar_extractor_step_decide_destination: destination %s", destination_name);
 
-  g_file_make_directory_with_parents (self->destination_dir,
+  g_file_make_directory_with_parents (new_destination,
                                       self->cancellable,
                                       &(self->error));
 
   if (g_error_matches (self->error, G_IO_ERROR, G_IO_ERROR_EXISTS)) {
     GFileType file_type;
 
-    file_type = g_file_query_file_type (self->destination_dir,
+    file_type = g_file_query_file_type (new_destination,
                                         G_FILE_QUERY_INFO_NONE,
                                         NULL);
 
@@ -1960,7 +1882,7 @@ autoar_extractor_step_extract (AutoarExtractor *self) {
     g_autoptr (GFile) extracted_filename = NULL;
     g_autoptr (GFile) hardlink_filename = NULL;
     AutoarConflictAction action;
-    gboolean file_conflict;
+    g_autoptr (GFile) file_conflict = NULL;
 
     if (g_cancellable_is_cancelled (self->cancellable)) {
       archive_read_free (a);
@@ -1972,34 +1894,53 @@ autoar_extractor_step_extract (AutoarExtractor *self) {
 
     extracted_filename =
       autoar_extractor_do_sanitize_pathname (self, pathname);
-    if (extracted_filename == NULL) {
-      archive_read_data_skip (a);
-      continue;
-    }
 
     if (hardlink != NULL) {
       hardlink_filename =
         autoar_extractor_do_sanitize_pathname (self, hardlink);
-        if (hardlink_filename == NULL) {
-          archive_read_data_skip (a);
-          continue;
-        }
     }
 
     /* Attempt to solve any name conflict before doing any operations */
-    file_conflict = autoar_extractor_check_file_conflict (extracted_filename,
+    file_conflict = autoar_extractor_check_file_conflict (self,
+                                                          extracted_filename,
                                                           archive_entry_filetype (entry));
     while (file_conflict) {
       GFile *new_extracted_filename = NULL;
 
+      /* Do not try to solve any conflicts in parents for now. Especially
+       * symlinks in parents are dangerous as it can easily happen that files
+       * are written outside of the destination. The tar cmd fails to extract
+       * such archives with ENOTDIR. Let's do the same here. This is most
+       * probably malicious, or corrupted archive if the conflict was caused
+       * only by files from the archive...
+       */
+      if (!g_file_equal (file_conflict, extracted_filename)) {
+        self->error = g_error_new (G_IO_ERROR,
+                                   G_IO_ERROR_NOT_DIRECTORY,
+                                   "The file is not a directory");
+        archive_read_free (a);
+        return;
+      }
+
       action = autoar_extractor_signal_conflict (self,
                                                  extracted_filename,
                                                  &new_extracted_filename);
 
       switch (action) {
         case AUTOAR_CONFLICT_OVERWRITE:
+          /* It is expected that this will fail for non-empty directories to
+           * prevent data loss.
+           */
+          g_file_delete (extracted_filename, self->cancellable, &self->error);
+          if (self->error != NULL) {
+            archive_read_free (a);
+            return;
+          }
           break;
         case AUTOAR_CONFLICT_CHANGE_DESTINATION:
+          /* FIXME: If the destination is changed for directory, it should be
+           * changed also for its children...
+           */
           g_assert_nonnull (new_extracted_filename);
           g_clear_object (&extracted_filename);
           extracted_filename = new_extracted_filename;
@@ -2016,11 +1957,15 @@ autoar_extractor_step_extract (AutoarExtractor *self) {
         break;
       }
 
-      file_conflict = autoar_extractor_check_file_conflict (extracted_filename,
+      g_clear_object (&file_conflict);
+      file_conflict = autoar_extractor_check_file_conflict (self,
+                                                            extracted_filename,
                                                             archive_entry_filetype (entry));
     }
 
     if (file_conflict && action == AUTOAR_CONFLICT_SKIP) {
+      self->total_files -= 1;
+      self->total_size -= archive_entry_size (entry);
       continue;
     }
 
