@@ -198,7 +198,7 @@ conflict_handler (AutoarExtractor *extractor,
                   gpointer user_data)
 {
   ExtractTestData *data = user_data;
-  AutoarConflictAction action = AUTOAR_CONFLICT_SKIP;
+  AutoarConflictAction action = AUTOAR_CONFLICT_UNHANDLED;
   gpointer value;
   gboolean key_found;
 
@@ -393,7 +393,8 @@ scan_directory (GFile *directory,
     if (g_file_info_get_file_type (file_info) == G_FILE_TYPE_DIRECTORY) {
       enumerator = g_file_enumerate_children (file,
                                               G_FILE_ATTRIBUTE_STANDARD_NAME","
-                                              G_FILE_ATTRIBUTE_STANDARD_TYPE,
+                                              G_FILE_ATTRIBUTE_STANDARD_TYPE","
+                                              G_FILE_ATTRIBUTE_STANDARD_SIZE,
                                               G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS,
                                               NULL, NULL);
 
@@ -459,7 +460,9 @@ reference_file_scanned (GFile *scanned_file,
     g_assert_cmpuint (g_file_info_get_file_type (scanned_file_info),
                       ==,
                       g_file_info_get_file_type (corresponding_file_info));
-
+    g_assert_cmpuint (g_file_info_get_size (scanned_file_info),
+                      ==,
+                      g_file_info_get_size (corresponding_file_info));
     g_hash_table_remove (extract_test->unmatched_files, relative_path);
   }
 }
@@ -556,6 +559,49 @@ test_one_file_different_name (void)
   assert_reference_and_output_match (extract_test);
 }
 
+/* Be sure that extra folder is not created in case of output-is-dest. */
+static void
+test_one_file_different_name_output_is_dest (void)
+{
+  /* arextract.zip
+   * └── arextractdifferent.txt
+   *
+   * 0 directories, 1 file
+   *
+   *
+   * ref
+   * └── arextractdifferent.txt
+   *
+   * 0 directory, 1 file
+   */
+
+  g_autoptr (ExtractTest) extract_test = NULL;
+  g_autoptr (ExtractTestData) data = NULL;
+  g_autoptr (GFile) archive = NULL;
+  g_autoptr (AutoarExtractor) extractor = NULL;
+
+  extract_test = extract_test_new ("test-one-file-different-name-output-is-dest");
+
+  if (!extract_test) {
+    g_assert_nonnull (extract_test);
+    return;
+  }
+
+  archive = g_file_get_child (extract_test->input, "arextract.zip");
+
+  extractor = autoar_extractor_new (archive, extract_test->output);
+  autoar_extractor_set_output_is_dest (extractor, TRUE);
+
+  data = extract_test_data_new_for_extract (extractor);
+
+  autoar_extractor_start (extractor, data->cancellable);
+
+  g_assert_cmpuint (data->number_of_files, ==, 1);
+  g_assert_no_error (data->error);
+  g_assert_true (data->completed_signalled);
+  assert_reference_and_output_match (extract_test);
+}
+
 static void
 test_multiple_files_same_name (void)
 {
@@ -654,7 +700,7 @@ test_multiple_files_different_name (void)
 }
 
 static void
-test_one_file_conflict_overwrite (void)
+test_conflict_overwrite (void)
 {
   /* arextract.zip
    * └── arextract.txt
@@ -675,7 +721,7 @@ test_one_file_conflict_overwrite (void)
   g_autoptr (GFile) reference_file = NULL;
   g_autoptr (AutoarExtractor) extractor = NULL;
 
-  extract_test = extract_test_new ("test-one-file-conflict-overwrite");
+  extract_test = extract_test_new ("test-conflict-overwrite");
 
   if (!extract_test) {
     g_assert_nonnull (extract_test);
@@ -687,8 +733,8 @@ test_one_file_conflict_overwrite (void)
   conflict_file = g_file_get_child (extract_test->output,
                                     "arextract.txt");
 
-  g_file_copy (reference_file, conflict_file, G_FILE_COPY_NONE,
-               NULL, NULL, NULL, NULL);
+  g_file_replace_contents (conflict_file, "this file should be overwritten", 31,
+                           NULL, FALSE, G_FILE_CREATE_NONE, NULL, NULL, NULL);
 
   archive = g_file_get_child (extract_test->input, "arextract.zip");
 
@@ -696,6 +742,10 @@ test_one_file_conflict_overwrite (void)
 
   data = extract_test_data_new_for_extract (extractor);
 
+  g_hash_table_insert (data->conflict_files_actions,
+                       g_object_ref (conflict_file),
+                       GUINT_TO_POINTER (AUTOAR_CONFLICT_OVERWRITE));
+
   autoar_extractor_start (extractor, data->cancellable);
 
   g_assert_cmpuint (data->number_of_files, ==, 1);
@@ -706,8 +756,173 @@ test_one_file_conflict_overwrite (void)
   assert_reference_and_output_match (extract_test);
 }
 
+/* Be sure that nonempty directories are not replaced to prevent data-loss. */
 static void
-test_one_file_conflict_new_destination (void)
+test_conflict_overwrite_nonempty_directory (void)
+{
+  /* arextract.zip
+   * └── arextract
+   *
+   * 0 directories, 1 file
+   *
+   *
+   * ref
+   * └── arextract
+   *     └── dummy
+   *
+   * 1 directory, 1 files
+   */
+
+  g_autoptr (ExtractTest) extract_test = NULL;
+  g_autoptr (ExtractTestData) data = NULL;
+  g_autoptr (GFile) archive = NULL;
+  g_autoptr (GFile) conflict_directory = NULL;
+  g_autoptr (GFile) dummy_file = NULL;
+  g_autoptr (GFileOutputStream) out = NULL;
+  g_autoptr (AutoarExtractor) extractor = NULL;
+
+  extract_test = extract_test_new ("test-conflict-overwrite-nonempty-directory");
+
+  if (!extract_test) {
+    g_assert_nonnull (extract_test);
+    return;
+  }
+
+  conflict_directory = g_file_get_child (extract_test->output,
+                                         "arextract");
+  dummy_file = g_file_get_child (conflict_directory, "dummy");
+
+  g_file_make_directory (conflict_directory, NULL, NULL);
+
+  out = g_file_create (dummy_file, G_FILE_CREATE_NONE, NULL, NULL);
+  g_output_stream_close (G_OUTPUT_STREAM (out), NULL, NULL);
+
+  archive = g_file_get_child (extract_test->input, "arextract.zip");
+
+  extractor = autoar_extractor_new (archive, extract_test->output);
+
+  data = extract_test_data_new_for_extract (extractor);
+
+  g_hash_table_insert (data->conflict_files_actions,
+                       g_object_ref (conflict_directory),
+                       GUINT_TO_POINTER (AUTOAR_CONFLICT_OVERWRITE));
+
+  autoar_extractor_start (extractor, data->cancellable);
+
+  g_assert_cmpuint (data->number_of_files, ==, 1);
+  g_assert_true (g_hash_table_contains (data->conflict_files,
+                                        conflict_directory));
+  g_assert_error (data->error, G_IO_ERROR, G_IO_ERROR_NOT_EMPTY);
+  g_assert_false (data->completed_signalled);
+  assert_reference_and_output_match (extract_test);
+}
+
+/* Be sure that symlink itself is replaced, not its target. */
+static void
+test_conflict_overwrite_symlink (void)
+{
+  /* arextract.tar
+   * ├── arextract -> arectract.txt
+   * └── arextract
+   *
+   * 0 directories, 2 files
+   *
+   *
+   * ref
+   * └── arextract
+   *
+   * 0 directories, 1 file
+   */
+
+  g_autoptr (ExtractTest) extract_test = NULL;
+  g_autoptr (ExtractTestData) data = NULL;
+  g_autoptr (GFile) archive = NULL;
+  g_autoptr (GFile) conflict_directory = NULL;
+  g_autoptr (GFile) conflict_file = NULL;
+  g_autoptr (AutoarExtractor) extractor = NULL;
+
+  extract_test = extract_test_new ("test-conflict-overwrite-symlink");
+
+  if (!extract_test) {
+    g_assert_nonnull (extract_test);
+    return;
+  }
+
+  conflict_file = g_file_get_child (extract_test->output, "arextract");
+  archive = g_file_get_child (extract_test->input, "arextract.tar");
+
+  extractor = autoar_extractor_new (archive, extract_test->output);
+
+  data = extract_test_data_new_for_extract (extractor);
+
+  g_hash_table_insert (data->conflict_files_actions,
+                       g_object_ref (conflict_file),
+                       GUINT_TO_POINTER (AUTOAR_CONFLICT_OVERWRITE));
+
+  autoar_extractor_start (extractor, data->cancellable);
+
+  g_assert_cmpuint (data->number_of_files, ==, 2);
+  g_assert_true (g_hash_table_contains (data->conflict_files,
+                                        conflict_file));
+  g_assert_true (data->completed_signalled);
+  assert_reference_and_output_match (extract_test);
+}
+
+/* Be sure that hardlink itself is replaced, not its target. */
+static void
+test_conflict_overwrite_hardlink (void)
+{
+  /* arextract.tar
+   * ├── arectract.txt
+   * ├── arextract -> arectract.txt
+   * └── arextract
+   *
+   * 0 directories, 3 files
+   *
+   *
+   * ref
+   * ├── arextract.txt
+   * └── arextract
+   *
+   * 0 directories, 2 files
+   */
+
+  g_autoptr (ExtractTest) extract_test = NULL;
+  g_autoptr (ExtractTestData) data = NULL;
+  g_autoptr (GFile) archive = NULL;
+  g_autoptr (GFile) conflict_file = NULL;
+  g_autoptr (AutoarExtractor) extractor = NULL;
+
+  extract_test = extract_test_new ("test-conflict-overwrite-hardlink");
+
+  if (!extract_test) {
+    g_assert_nonnull (extract_test);
+    return;
+  }
+
+  conflict_file = g_file_get_child (extract_test->output, "arextract");
+  archive = g_file_get_child (extract_test->input, "arextract.tar");
+
+  extractor = autoar_extractor_new (archive, extract_test->output);
+  autoar_extractor_set_output_is_dest (extractor, TRUE);
+
+  data = extract_test_data_new_for_extract (extractor);
+
+  g_hash_table_insert (data->conflict_files_actions,
+                       g_object_ref (conflict_file),
+                       GUINT_TO_POINTER (AUTOAR_CONFLICT_OVERWRITE));
+
+  autoar_extractor_start (extractor, data->cancellable);
+
+  g_assert_cmpuint (data->number_of_files, ==, 3);
+  g_assert_true (g_hash_table_contains (data->conflict_files,
+                                        conflict_file));
+  g_assert_true (data->completed_signalled);
+  assert_reference_and_output_match (extract_test);
+}
+
+static void
+test_conflict_new_destination (void)
 {
   /* arextract.zip
    * └── arextract.txt
@@ -729,7 +944,7 @@ test_one_file_conflict_new_destination (void)
   g_autoptr (GFile) reference_file = NULL;
   g_autoptr (AutoarExtractor) extractor = NULL;
 
-  extract_test = extract_test_new ("test-one-file-conflict-new-destination");
+  extract_test = extract_test_new ("test-conflict-new-destination");
 
   if (!extract_test) {
     g_assert_nonnull (extract_test);
@@ -770,7 +985,7 @@ test_one_file_conflict_new_destination (void)
 }
 
 static void
-test_one_file_conflict_skip_file (void)
+test_conflict_skip_file (void)
 {
   /* arextract.zip
    * └── arextract.txt
@@ -791,7 +1006,7 @@ test_one_file_conflict_skip_file (void)
   g_autoptr (GFile) reference_file = NULL;
   g_autoptr (AutoarExtractor) extractor = NULL;
 
-  extract_test = extract_test_new ("test-one-file-conflict-skip-file");
+  extract_test = extract_test_new ("test-conflict-skip-file");
 
   if (!extract_test) {
     g_assert_nonnull (extract_test);
@@ -826,48 +1041,43 @@ test_one_file_conflict_skip_file (void)
   assert_reference_and_output_match (extract_test);
 }
 
+/* Be sure that the default action is skip to prevent data-loss. */
 static void
-test_one_file_error_file_over_directory (void)
+test_conflict_skip_file_default (void)
 {
-  /* The dummy file in this test is not relevant to the test itself, but it
-   * was required in order to add the directory to the .git repo
-   *
-   * arextract.zip
-   * └── arextract
+  /* arextract.zip
+   * └── arextract.txt
    *
    * 0 directories, 1 file
    *
    *
    * ref
-   * └── arextract
-   *     └── dummy
+   * └── arextract.txt
    *
-   * 1 directory, 1 files
+   * 0 directories, 1 file
    */
 
   g_autoptr (ExtractTest) extract_test = NULL;
   g_autoptr (ExtractTestData) data = NULL;
   g_autoptr (GFile) archive = NULL;
-  g_autoptr (GFile) conflict_directory = NULL;
-  g_autoptr (GFile) dummy_file = NULL;
-  g_autoptr (GFileOutputStream) out = NULL;
+  g_autoptr (GFile) conflict_file = NULL;
+  g_autoptr (GFile) reference_file = NULL;
   g_autoptr (AutoarExtractor) extractor = NULL;
 
-  extract_test = extract_test_new ("test-one-file-error-file-over-directory");
+  extract_test = extract_test_new ("test-conflict-skip-file-default");
 
   if (!extract_test) {
     g_assert_nonnull (extract_test);
     return;
   }
 
-  conflict_directory = g_file_get_child (extract_test->output,
-                                         "arextract");
-  dummy_file = g_file_get_child (conflict_directory, "dummy");
+  reference_file = g_file_get_child (extract_test->reference,
+                                     "arextract.txt");
+  conflict_file = g_file_get_child (extract_test->output,
+                                    "arextract.txt");
 
-  g_file_make_directory (conflict_directory, NULL, NULL);
-
-  out = g_file_create (dummy_file, G_FILE_CREATE_NONE, NULL, NULL);
-  g_output_stream_close (G_OUTPUT_STREAM (out), NULL, NULL);
+  g_file_copy (reference_file, conflict_file, G_FILE_COPY_NONE,
+               NULL, NULL, NULL, NULL);
 
   archive = g_file_get_child (extract_test->input, "arextract.zip");
 
@@ -875,12 +1085,16 @@ test_one_file_error_file_over_directory (void)
 
   data = extract_test_data_new_for_extract (extractor);
 
+  g_hash_table_insert (data->conflict_files_actions,
+                       g_object_ref (conflict_file),
+                       GUINT_TO_POINTER (AUTOAR_CONFLICT_UNHANDLED));
+
   autoar_extractor_start (extractor, data->cancellable);
 
   g_assert_cmpuint (data->number_of_files, ==, 1);
   g_assert_true (g_hash_table_contains (data->conflict_files,
-                                        conflict_directory));
-  g_assert_null (data->error);
+                                        conflict_file));
+  g_assert_no_error (data->error);
   g_assert_true (data->completed_signalled);
   assert_reference_and_output_match (extract_test);
 }
@@ -934,6 +1148,144 @@ test_change_extract_destination (void)
   assert_reference_and_output_match (extract_test);
 }
 
+/* Be sure that the files with symlinks in parents are refused completely for
+ * security reasons.
+ *
+ * If symlinks in parents will be allowed in the future, then the following
+ * cases need to be tested to be sure that nothing is written outside of the
+ * destination:
+ * 1) link -> .. ; link/arextract.txt
+ * 2) link -> /tmp ; link/arextract.txt
+ * 3) current -> . ; link -> current/.. ; link/arextract.txt
+ * 4) tmplink -> /tmp ; link -> tmplink/.. ; link/arextract.txt
+ */
+static void
+test_symlink_parent (void)
+{
+  /* arextract.tar
+   * ├── arextract -> /tmp
+   * └── arextract/arextract.txt
+   *
+   * 0 directories, 2 files
+   *
+   *
+   * ref
+   * └── arextract -> /tmp
+   *
+   * 0 directories, 1 file
+   */
+
+  g_autoptr (ExtractTest) extract_test = NULL;
+  g_autoptr (ExtractTestData) data = NULL;
+  g_autoptr (GFile) archive = NULL;
+  g_autoptr (AutoarExtractor) extractor = NULL;
+
+  extract_test = extract_test_new ("test-symlink-parent");
+
+  if (!extract_test) {
+    g_assert_nonnull (extract_test);
+    return;
+  }
+
+  archive = g_file_get_child (extract_test->input, "arextract.tar");
+
+  extractor = autoar_extractor_new (archive, extract_test->output);
+
+  data = extract_test_data_new_for_extract (extractor);
+
+  autoar_extractor_start (extractor, data->cancellable);
+
+  g_assert_cmpuint (data->number_of_files, ==, 2);
+  g_assert_error (data->error, G_IO_ERROR, G_IO_ERROR_NOT_DIRECTORY);
+  g_assert_false (data->completed_signalled);
+  assert_reference_and_output_match (extract_test);
+}
+
+/* Be sure that file with the ".." parent is written in the destination. */
+static void
+test_sanitize_dotdot_parent (void)
+{
+  /* arextract.tar
+   * └── ./../arextract.txt
+   *
+   * 0 directories, 1 file
+   *
+   *
+   * ref
+   * └── arextract.txt
+   *
+   * 0 directories, 1 file
+   */
+
+  g_autoptr (ExtractTest) extract_test = NULL;
+  g_autoptr (ExtractTestData) data = NULL;
+  g_autoptr (GFile) archive = NULL;
+  g_autoptr (AutoarExtractor) extractor = NULL;
+
+  extract_test = extract_test_new ("test-sanitize-dotdot-parent");
+
+  if (!extract_test) {
+    g_assert_nonnull (extract_test);
+    return;
+  }
+
+  archive = g_file_get_child (extract_test->input, "arextract.tar");
+
+  extractor = autoar_extractor_new (archive, extract_test->output);
+  autoar_extractor_set_output_is_dest (extractor, TRUE);
+
+  data = extract_test_data_new_for_extract (extractor);
+
+  autoar_extractor_start (extractor, data->cancellable);
+
+  g_assert_cmpuint (data->number_of_files, ==, 1);
+  g_assert_no_error (data->error);
+  g_assert_true (data->completed_signalled);
+  assert_reference_and_output_match (extract_test);
+}
+
+/* Be sure that the absolute paths are relative to the destination. */
+static void
+test_sanitize_absolute_path (void)
+{
+  /* arextract.tar
+   * └── /arextract.txt
+   *
+   * 0 directories, 1 file
+   *
+   *
+   * ref
+   * └── arextract.txt
+   *
+   * 0 directories, 1 file
+   */
+
+  g_autoptr (ExtractTest) extract_test = NULL;
+  g_autoptr (ExtractTestData) data = NULL;
+  g_autoptr (GFile) archive = NULL;
+  g_autoptr (AutoarExtractor) extractor = NULL;
+
+  extract_test = extract_test_new ("test-sanitize-absolute-path");
+
+  if (!extract_test) {
+    g_assert_nonnull (extract_test);
+    return;
+  }
+
+  archive = g_file_get_child (extract_test->input, "arextract.tar");
+
+  extractor = autoar_extractor_new (archive, extract_test->output);
+
+  data = extract_test_data_new_for_extract (extractor);
+
+  autoar_extractor_start (extractor, data->cancellable);
+
+  g_assert_cmpuint (data->number_of_files, ==, 1);
+  g_assert_no_error (data->error);
+  g_assert_true (data->completed_signalled);
+  assert_reference_and_output_match (extract_test);
+}
+
 static void
 setup_test_suite (void)
 {
@@ -941,23 +1293,37 @@ setup_test_suite (void)
                    test_one_file_same_name);
   g_test_add_func ("/autoar-extract/test-one-file-different-name",
                    test_one_file_different_name);
+  g_test_add_func ("/autoar-extract/test-one-file-different-name-output-is-dest",
+                   test_one_file_different_name_output_is_dest);
   g_test_add_func ("/autoar-extract/test-multiple-files-same-name",
                    test_multiple_files_same_name);
   g_test_add_func ("/autoar-extract/test-multiple-files-different-name",
                    test_multiple_files_different_name);
 
-  g_test_add_func ("/autoar-extract/test-one-file-conflict-overwrite",
-                   test_one_file_conflict_overwrite);
-  g_test_add_func ("/autoar-extract/test-one-file-conflict-new-destination",
-                   test_one_file_conflict_new_destination);
-  g_test_add_func ("/autoar-extract/test-one-file-conflict-skip-file",
-                   test_one_file_conflict_skip_file);
-
-  g_test_add_func ("/autoar-extract/test-one-file-error-file-over-directory",
-                   test_one_file_error_file_over_directory);
+  g_test_add_func ("/autoar-extract/test-conflict-overwrite",
+                   test_conflict_overwrite);
+  g_test_add_func ("/autoar-extract/test-conflict-overwrite-nonempty-directory",
+                   test_conflict_overwrite_nonempty_directory);
+  g_test_add_func ("/autoar-extract/test-conflict-overwrite-symlink",
+                   test_conflict_overwrite_symlink);
+  g_test_add_func ("/autoar-extract/test-conflict-overwrite-hardlink",
+                   test_conflict_overwrite_hardlink);
+  g_test_add_func ("/autoar-extract/test-conflict-new-destination",
+                   test_conflict_new_destination);
+  g_test_add_func ("/autoar-extract/test-conflict-skip-file",
+                   test_conflict_skip_file);
+  g_test_add_func ("/autoar-extract/test-conflict-skip-file-default",
+                   test_conflict_skip_file_default);
 
   g_test_add_func ("/autoar-extract/test-change-extract-destination",
                    test_change_extract_destination);
+
+  g_test_add_func ("/autoar-extract/test-symlink-parent",
+                   test_symlink_parent);
+  g_test_add_func ("/autoar-extract/test-sanitize-dotdot-parent",
+                   test_sanitize_dotdot_parent);
+  g_test_add_func ("/autoar-extract/test-sanitize-absolute-path",
+                   test_sanitize_absolute_path);
 }
 
 int
